/*
 *   Copyright (C) 1988-2006 Lucent Technologies Inc. and others.
 *   All Rights Reserved.
 */

/*	cscope - interactive C symbol cross-reference
 *
 *	keyword look-up routine for the Java symbol scanner
 */

#ifdef MULTILINGUAL

#include "global.h"

/* keyword text for fast testing of keywords in the scanner */
char	classtext[] = "class";
char	interfacetext[] = "interface";

/* This keyword table is also used for keyword text compression.  Keywords
 * with an index less than the numeric value of a space are replaced with the
 * control character corresponding to the index, so they cannot be moved
 * without changing the database file version and adding compatibility code
 * for old databases.
 */
struct	keystruct jkeyword[] = {
	{ "#define",	' ',	MISC,	NULL },	/* must be table entry 0 for old databases */
	{ "boolean",	' ',	DECL,	NULL },
	{ "break",	'\0',	FLOW,	NULL },
	{ "byte",	' ',	DECL,	NULL },
	{ "case",	' ',	FLOW,	NULL },
	{ "catch",	'(',	MISC,	NULL },
	{ "char",	' ',	DECL,	NULL },
	{ classtext,	' ',	DECL,	NULL },
	{ "continue",	'\0',	FLOW,	NULL },
	{ "\t",		'\0',	MISC,	NULL },	/* must be table entry 9 */
	{ "\n",		'\0',	MISC,	NULL },	/* must be table entry 10 */
	{ "default",	'\0',	FLOW,	NULL },
	{ "do",		' ',	FLOW,	NULL },
	{ "double",	' ',	DECL,	NULL },
	{ "else",	' ',	FLOW,	NULL },
	{ "final",	' ',	MISC,	NULL },
	{ "float",	' ',	DECL,	NULL },
	{ "for",	'(',	FLOW,	NULL },
	{ "if",		'(',	FLOW,	NULL },
	{ "import",	' ',	MISC,	NULL },
	{ "int",	' ',	DECL,	NULL }, /* 20 */
	{ "long",	' ',	DECL,	NULL },
	{ "private",	' ',	DECL,	NULL },
	{ "public",	' ',	DECL,	NULL },
	{ "return",	'\0',	FLOW,	NULL },
	{ "short",	' ',	DECL,	NULL },
	{ "static",	' ',	DECL,	NULL },
	{ "switch",	'(',	FLOW,	NULL },
	{ "synchronized", ' ',	MISC,	NULL },
	{ "try",	' ',	MISC,	NULL },
	{ "void",	' ',	DECL,	NULL }, /* 30 */
	{ "while",	'(',	FLOW,	NULL },
	
	/* these keywords are not compressed */
	{ "abstract",	' ',	MISC,	NULL },
	{ "const",	' ',	MISC,	NULL },
	{ "extends",	' ',	MISC,	NULL },
	{ "false",	' ',	MISC,	NULL },
	{ "finally",	' ',	MISC,	NULL },
	{ "goto",	' ',	MISC,	NULL },
	{ "implements",	' ',	MISC,	NULL },
	{ "instanceof",	' ',	MISC,	NULL },
	{ interfacetext,' ',	MISC,	NULL },
	{ "native",	' ',	MISC,	NULL },
	{ "new",	' ',	MISC,	NULL },
	{ "null",	' ',	MISC,	NULL },
	{ "package",	' ',	MISC,	NULL },
	{ "protected",	' ',	MISC,	NULL },
	{ "strictfp",	' ',	MISC,	NULL },
	{ "super",	' ',	MISC,	NULL },
	{ "this",	' ',	MISC,	NULL },
	{ "throw",	' ',	MISC,	NULL },
	{ "throws",	' ',	MISC,	NULL },
	{ "transient",	' ',	DECL,	NULL },
	{ "true",	' ',	MISC,	NULL },
	{ "volatile",	' ',	DECL,	NULL },
};
#define KEYWORDS	(sizeof(jkeyword) / sizeof(struct keystruct))

#define HASHMOD	(KEYWORDS * 2 + 1)

static	struct	keystruct *hashtab[HASHMOD]; /* pointer table */

#if defined( __STDC__ ) || defined( __BORLANDC__ )
void	jinitsymtab(void);
struct	keystruct *jlookup(char *ident);
#endif

/* put the keywords into the symbol table */

void
jinitsymtab()
{
	register int	i, j;
	register struct	keystruct *p;
	
	for (i = 1; i < KEYWORDS; ++i) {
		p = &jkeyword[i];
		j = hash(p->text) % HASHMOD;
		p->next = hashtab[j];
		hashtab[j] = p;
	}
}

/* see if this identifier is a keyword */

struct	keystruct *
jlookup(ident)
register char	*ident;
{
	register struct	keystruct *p;
	int	c;
	
	/* look up the identifier in the keyword table */
	for (p = hashtab[hash(ident) % HASHMOD]; p != NULL; p = p->next) {
		if (strequal(ident, p->text)) {
			if (compress == YES && (c = p - jkeyword) < ' ') {
				ident[0] = c;	/* compress the keyword */
			}
			return(p);
		}
	}
	/* this is an identifier */
	return(NULL);
}

#endif
